import inspect
from copy import deepcopy
from types import FunctionType
from typing import Callable, Type

import pydantic


def copy_function(fn) -> Callable:
    """
    Create a clone of a given function
    :param fn: Function to copy
    :return: New function with the same attributes but with different id
    """
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    g = FunctionType(
        fn.__code__,
        fn.__globals__,
        name=fn.__name__,
        argdefs=fn.__defaults__,
        closure=fn.__closure__,
    )
    g.__kwdefaults__ = deepcopy(fn.__kwdefaults__)
    g.__annotations__ = deepcopy(fn.__annotations__)
    return g


def add_annotation_to_first_argument(fn: FunctionType, model: Type[pydantic.BaseModel]):
    """
    Patch given function so its first argument without type annotation
    becomes type-hinted with model
    :param fn: Function to patch
    :param model: Type to add to the first argument
    """
    fn_spec = inspect.getfullargspec(fn)
    if not len(fn_spec.args):
        raise ValueError(f"Function {fn.__name__} has no arguments")
    untyped_args = [a for a in fn_spec.args if a not in fn.__annotations__]
    if untyped_args:
        fn.__annotations__[untyped_args[0]] = model



from pathlib import Path

from .core import BaseValidator, OpenApiValidationError


class IhanStandardError(OpenApiValidationError):
    pass


class WrongContentType(IhanStandardError):
    pass


class SchemaMissing(IhanStandardError):
    pass


class NoEndpointsDefined(IhanStandardError):
    pass


class OnlyOneEndpointAllowed(IhanStandardError):
    pass


class PostMethodIsMissing(IhanStandardError):
    pass


class OnlyPostMethodAllowed(IhanStandardError):
    pass


class RequestBodyMissing(IhanStandardError):
    pass


class ResponseBodyMissing(IhanStandardError):
    pass


class AuthorizationHeaderMissing(IhanStandardError):
    pass


class AuthProviderHeaderMissing(IhanStandardError):
    pass


class StandardComponentMissing(IhanStandardError):
    pass


class StandardContentMissing(IhanStandardError):
    pass


class ServersShouldNotBeDefined(IhanStandardError):
    pass


class SecurityShouldNotBeDefined(IhanStandardError):
    pass


def validate_component_schema(spec: dict, components_schema: dict):
    if not spec["content"].get("application/json"):
        raise WrongContentType("Model description must be in application/json format")
    ref = spec["content"]["application/json"].get("schema", {}).get("$ref")
    if not ref:
        raise SchemaMissing(
            'Request or response model is missing from "schema/$ref" section'
        )
    if not ref.startswith("#/components/schemas/"):
        raise SchemaMissing(
            "Request and response models must be defined at"
            '"#/components/schemas/" section'
        )
    model_name = ref.split("/")[-1]
    if not components_schema.get(model_name):
        raise SchemaMissing(f"Component schema is missed for {model_name}")


def validate_spec(spec: dict):
    """
    Validate that OpenAPI spec looks like a standard. For example, that
    it contains only one POST method defined.

    :param spec: OpenAPI spec
    :raises OpenApiValidationError: When OpenAPI spec is incorrect
    """
    if "servers" in spec:
        raise ServersShouldNotBeDefined('"servers" section found')

    paths = spec.get("paths", {})
    if not paths:
        raise NoEndpointsDefined
    if len(paths) > 1:
        raise OnlyOneEndpointAllowed

    post_route = {}
    for name, path in paths.items():
        methods = list(path)
        if "post" not in methods:
            raise PostMethodIsMissing
        if methods != ["post"]:
            raise OnlyPostMethodAllowed
        post_route = path["post"]

    component_schemas = spec.get("components", {}).get("schemas")
    if not component_schemas:
        raise SchemaMissing('No "components/schemas" section defined')

    if "security" in post_route:
        raise SecurityShouldNotBeDefined('"security" section found')

    if post_route.get("requestBody", {}).get("content"):
        validate_component_schema(post_route["requestBody"], component_schemas)

    responses = post_route.get("responses", {})
    if not responses.get("200") or not responses["200"].get("content"):
        raise ResponseBodyMissing
    validate_component_schema(responses["200"], component_schemas)

    headers = [
        param.get("name", "").lower()
        for param in post_route.get("parameters", [])
        if param.get("in") == "header"
    ]
    if "authorization" not in headers:
        raise AuthorizationHeaderMissing
    if "x-authorization-provider" not in headers:
        raise AuthProviderHeaderMissing


def check_extra_files_exist(path: Path):
    html = path.with_suffix(".html")
    if not html.exists():
        raise StandardComponentMissing(f"Missing {html}")
    if html.read_text().strip() == "":
        raise StandardContentMissing(f"Make sure {html} is not empty")


class IhanStandardsValidator(BaseValidator):
    def validate_spec(self, spec: dict):
        # just to reduce indentation
        return validate_spec(spec)

    def validate(self):
        super().validate()
        check_extra_files_exist(self.path)



import importlib.util
import logging
import sys
import traceback
import uuid
from contextlib import contextmanager
from pathlib import Path
from typing import List

import click
import coloredlogs

from openapi_to_fastapi.validator.core import BaseValidator, DefaultValidator

from .routes import SpecRouter
from .validator import ihan_standards

logger = logging.getLogger("openapi_to_fastapi_cli")
coloredlogs.install(logger=logger, fmt="%(message)s")


def print_dashes(char="=", length=79):
    logger.info(char * length)


@contextmanager
def header():
    print_dashes()
    yield
    print_dashes()


@click.command()
@click.option(
    "--path", "-p", required=True, help="Path to directory with OpenAPI specs"
)
@click.option("--validator", "-v", multiple=True, help="Class name of extra validator")
@click.option(
    "--module",
    "-m",
    multiple=True,
    help="Module with extra validator class definitions",
)
def cli_validate_specs(path: str, validator, module):
    sys.exit(validate_specs(Path(path), module, validator))


def validate_specs(path: Path, modules: List[str], extra_validators: List[str]) -> int:
    validators = _load_validators(extra_validators, modules)
    with header():
        logger.info(f"OpenAPI specs root path: {path}")
        logger.info("Validators: %s", ", ".join([v.__name__ for v in validators]))

    passed, failed = 0, 0
    for spec_path in path.glob("**/*.json"):
        logger.info(f"File: {spec_path}")
        try:
            SpecRouter(spec_path, validators)
        except Exception:
            logger.error("\n%s", traceback.format_exc())
            logger.error("[FAILED]")
            failed += 1
        else:
            logger.info("[PASSED]")
            passed += 1
        print_dashes("-")

    with header():
        log = logger.error if failed else logger.info
        log("Summary:")
        log(f"Total : {passed+failed}")
        log(f"Passed: {passed}")
        log(f"Failed: {failed}")
    return 1 if failed else 0


def _load_extra_validator_modules(modules: List[str]) -> list:
    validator_modules = [ihan_standards]
    for module_path in modules:
        module_name = f"oas_models_{uuid.uuid4()}"
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        if spec.loader:
            validator_modules.append(spec.loader.load_module(module_name))
    return validator_modules


def _load_validators(
    validators: List[str],
    extra_modules: List[str],
) -> list:
    _validators = [DefaultValidator]

    for name in validators:
        for validator_module in _load_extra_validator_modules(extra_modules):
            validator = getattr(validator_module, name, None)
            if validator and issubclass(validator, BaseValidator):
                _validators.append(validator)
                break
        else:
            raise ValueError(f"Failed to load validator: {name}")
    return _validators



import importlib.util
import tempfile
import uuid

from datamodel_code_generator import (
    BaseModel,
    CustomRootType,
    DataModelField,
    PythonVersion,
)
from datamodel_code_generator.parser.openapi import OpenAPIParser

from openapi_to_fastapi.logger import logger


def generate_model_from_schema(schema: str) -> str:
    """
    Given an OpenAPI schema, generate pydantic models from everything defined
    in the "components/schemas" section

    :param schema: Content of an OpenAPI spec, plain text
    :return: Importable python code with generated models
    """
    parser = OpenAPIParser(
        BaseModel,
        CustomRootType,
        DataModelField,
        base_class="pydantic.BaseModel",
        custom_template_dir=None,
        extra_template_data=None,
        target_python_version=PythonVersion.PY_37,
        text=schema,
        dump_resolve_reference_action=None,
        validation=True,
        field_constraints=False,
        snake_case_field=False,
        strip_default_none=False,
        aliases=None,
    )

    result = parser.parse()
    return str(result)


def load_models(schema: str, name: str = "", cleanup: bool = True):
    """
    Generate pydantic models from OpenAPI spec and return a python module,
    which contains all the models from the "components/schemas" section.
    This function will create a dedicated python file in OS's temporary dir
    and imports it
    :param schema: OpenAPI spec, plain text
    :param name: Prefix for a module name, optional
    :param cleanup: Whether to remove a file with models afterwards
    :return: Module with pydantic models
    """
    prefix = name.replace("/", "").replace(" ", "").replace("\\", "") + "_"
    with tempfile.NamedTemporaryFile(
        prefix=prefix, mode="w", suffix=".py", delete=cleanup
    ) as tmp_file:
        model_py = generate_model_from_schema(schema)
        tmp_file.write(model_py)
        if not cleanup:
            logger.info("Generated module %s: %s", name, tmp_file.name)
        tmp_file.flush()
        module_name = f"oas_models_{uuid.uuid4()}"
        spec = importlib.util.spec_from_file_location(module_name, tmp_file.name)
        if spec.loader:
            return spec.loader.load_module(module_name)
        else:
            raise ValueError(f"Failed to load module {module_name}")



import abc
import json
from pathlib import Path
from typing import Union


class OpenApiValidationError(Exception):
    pass


class InvalidJSON(OpenApiValidationError):
    pass


class UnsupportedVersion(OpenApiValidationError):
    pass


class MissingParameter(OpenApiValidationError):
    pass


class BaseValidator:
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def validate(self):
        try:
            spec = json.loads(self.path.read_text())
        except json.JSONDecodeError:
            raise InvalidJSON(f"Incorrect JSON: {self.path}")
        self.validate_spec(spec)

    @abc.abstractmethod
    def validate_spec(self, spec: dict):
        raise NotImplementedError


class DefaultValidator(BaseValidator):
    def validate_spec(self, spec: dict):
        if not spec.get("openapi", "").startswith("3"):
            raise UnsupportedVersion



import json

import pydantic
import pytest
from fastapi import Header

from openapi_to_fastapi.model_generator import load_models
from openapi_to_fastapi.routes import SpecRouter

# values aligned with the response defined in the data/ihan/CompanyBasicInfo.json
company_basic_info_resp = {
    "name": "Company",
    "companyId": "test",
    "companyForm": "Form",
    "registrationDate": "Long ago",
}


def test_routes_are_created(ihan_client, specs_root):
    assert ihan_client.post("/Company/BasicInfo").status_code != 404
    assert ihan_client.post("/Non/Existing/Stuff").status_code == 404

    assert ihan_client.get("/Company/BasicInfo").status_code == 405
    assert ihan_client.get("/Non/Existing/Stuff").status_code == 404


def test_pydantic_model_loading(specs_root):
    path = specs_root / "ihan" / "CompanyBasicInfo.json"
    spec = json.loads(path.read_text())
    module = load_models(spec, "/Company/BasicInfo")
    assert module.BasicCompanyInfoRequest
    assert module.BasicCompanyInfoResponse
    assert module.HTTPValidationError
    assert module.ValidationError

    with pytest.raises(pydantic.ValidationError):
        module.BasicCompanyInfoRequest()

    with pytest.raises(pydantic.ValidationError):
        module.BasicCompanyInfoRequest(companyId=[])

    company_info_req = module.BasicCompanyInfoRequest(companyId="abc")
    assert company_info_req.dict() == {"companyId": "abc"}

    assert module.ValidationError(loc=[], msg="Crap", type="Error")


def test_weather_route_payload_errors(app, specs_root, client, snapshot):
    spec_router = SpecRouter(specs_root / "ihan")

    @spec_router.post("/Weather/Current/Metric")
    def weather_metric(request):
        return {}

    app.include_router(spec_router.to_fastapi_router())

    resp = client.post("/Weather/Current/Metric", json={})
    assert resp.status_code == 422
    snapshot.assert_match(resp.json(), "Missing payload")

    resp = client.post("/Weather/Current/Metric", json={"lat": "1,1.2", "lon": "99999"})
    assert resp.status_code == 422
    snapshot.assert_match(resp.json(), "Incorrect payload type")


def test_company_custom_post_route(app, client, specs_root, snapshot):
    spec_router = SpecRouter(specs_root / "ihan")

    @spec_router.post("/Company/BasicInfo")
    def weather_metric(request):
        return company_basic_info_resp

    app.include_router(spec_router.to_fastapi_router())

    resp = client.post("/Company/BasicInfo", json={"companyId": "test"})
    assert resp.status_code == 200, resp.json()
    assert resp.json() == company_basic_info_resp


def test_default_post_handler(app, client, specs_root, snapshot):
    spec_router = SpecRouter(specs_root / "ihan")

    @spec_router.post()
    def company_info(request):
        return company_basic_info_resp

    app.include_router(spec_router.to_fastapi_router())
    resp = client.post("/Company/BasicInfo", json={"companyId": "test"})
    assert resp.status_code == 200
    assert resp.json() == company_basic_info_resp


def test_custom_route_definitions(app, client, specs_root, snapshot):
    spec_router = SpecRouter(specs_root / "ihan")

    @spec_router.post("/Weather/Current/Metric")
    def weather_metric(request, vendor: str, auth_header: str = Header(...)):
        return {"customRoute": ""}

    app.include_router(spec_router.to_fastapi_router())
    resp = client.post("/Weather/Current/Metric", json={"lat": "30.5", "lon": 1.56})
    assert resp.status_code == 422
    snapshot.assert_match(resp.json(), "Custom route definition")


def test_response_model_is_parsed(app, client, specs_root):
    spec_router = SpecRouter(specs_root / "ihan")

    @spec_router.post("/Weather/Current/Metric")
    def weather_metric(request):
        return {}

    spec_router.to_fastapi_router()
    assert spec_router.get_response_model("/Weather/Current/Metric", "get") is None
    resp_model = spec_router.get_response_model("/Weather/Current/Metric", "post")

    assert resp_model is not None
    with pytest.raises(pydantic.ValidationError):
        resp_model()

    assert resp_model(
        humidity=1, pressure=1, rain=False, temp=30, windSpeed=1, windDirection=1.0
    )


def test_routes_meta_info(app, client, specs_root):
    spec_router = SpecRouter(specs_root / "ihan")

    @spec_router.post(
        "/Weather/Current/Metric",
        name="The Route",
        tags=["Routes"],
        description="Route description",
        response_description="Response description",
    )
    def weather_metric(request):
        return {}

    router = spec_router.to_fastapi_router()
    route = [r for r in router.routes if r.path == "/Weather/Current/Metric"][0]
    assert route.name == "The Route"
    assert route.tags == ["Routes"]
    assert route.description == "Route description"
    assert route.response_description == "Response description"

    # test that generating OpenAPI still works
    app.include_router(router)
    assert app.openapi()


def test_routes_meta_info_custom_name(app, client, specs_root):
    spec_router = SpecRouter(specs_root / "ihan")

    def name_factory(path="", **kwargs):
        return path[1:]

    @spec_router.post("/Company/BasicInfo", name_factory=name_factory)
    def weather_metric(request):
        return {}

    router = spec_router.to_fastapi_router()
    route = [r for r in router.routes if r.path == "/Company/BasicInfo"][0]

    assert route.name == "Company/BasicInfo"
    # description by default is coming from spec
    assert route.description == "Information about the company"

    # test that generating OpenAPI still works
    app.include_router(router)
    assert app.openapi()


def test_custom_route_name_for_default_post(app, client, specs_root):
    spec_router = SpecRouter(specs_root / "ihan")

    def name_factory(path="", **kwargs):
        return path[1:]

    @spec_router.post(name_factory=name_factory)
    def weather_metric(request):
        return {}

    router = spec_router.to_fastapi_router()
    route = [r for r in router.routes if r.path == "/Company/BasicInfo"][0]
    assert route.name == "Company/BasicInfo"

    # test that generating OpenAPI still works
    app.include_router(router)
    assert app.openapi()



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_weather_route_payload_errors Missing payload"] = {
    "detail": [
        {
            "loc": ["body", "request", "lat"],
            "msg": "field required",
            "type": "value_error.missing",
        },
        {
            "loc": ["body", "request", "lon"],
            "msg": "field required",
            "type": "value_error.missing",
        },
    ]
}

snapshots["test_weather_route_payload_errors Incorrect payload type"] = {
    "detail": [
        {
            "loc": ["body", "request", "lat"],
            "msg": "value is not a valid float",
            "type": "type_error.float",
        },
        {
            "ctx": {"limit_value": 180.0},
            "loc": ["body", "request", "lon"],
            "msg": "ensure this value is less than or equal to 180.0",
            "type": "value_error.number.not_le",
        },
    ]
}

snapshots["test_custom_route_definitions Custom route definition"] = {
    "detail": [
        {
            "loc": ["query", "vendor"],
            "msg": "field required",
            "type": "value_error.missing",
        },
        {
            "loc": ["header", "auth-header"],
            "msg": "field required",
            "type": "value_error.missing",
        },
    ]
}



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Optional, Type, Union

import pydantic
from fastapi import APIRouter

from .model_generator import load_models
from .parser import parse_openapi_spec
from .utils import add_annotation_to_first_argument, copy_function
from .validator.core import BaseValidator, DefaultValidator


def dummy_route(request):
    """
    Default handler to use if nothing was provided by user
    :param request: Incoming request
    :return: Empty JSON response
    """
    return {}


class EmptyBody(pydantic.BaseModel):
    pass


@dataclass
class RouteInfo:
    description: Optional[str] = None
    name: Optional[str] = None
    name_factory: Optional[Callable] = None
    response_description: str = "Successful response"
    tags: Optional[List[str]] = None

    request_model: Optional[Type[pydantic.BaseModel]] = None
    response_model: Optional[Type[pydantic.BaseModel]] = None
    handler: Callable = dummy_route


@dataclass
class RoutesMapping:
    default_get: RouteInfo = RouteInfo()
    default_post: RouteInfo = RouteInfo()
    get_map: Optional[Dict[str, RouteInfo]] = None
    post_map: Optional[Dict[str, RouteInfo]] = None


class SpecRouter:
    def __init__(
        self,
        specs_path: Union[str, Path],
        validators: List[Type[BaseValidator]] = None,
    ):
        self._validators = [DefaultValidator] + (validators or [])  # type: ignore
        self._routes = RoutesMapping(post_map={}, get_map={})

        self.specs_path = Path(specs_path)
        self._validate_and_parse_specs()

    def _validate_and_parse_specs(self):
        """
        Validate OpenAPI specs and parse required information from them
        """
        if self.specs_path.is_file():
            specs = [self.specs_path]  # for CLI usage
        else:
            specs = self.specs_path.glob("**/*.json")

        for spec_path in specs:
            for validator in self._validators:
                validator(spec_path).validate()

            raw_spec = spec_path.read_text()
            json_spec = json.loads(raw_spec)
            for path, path_item in parse_openapi_spec(json_spec).items():
                models = load_models(raw_spec, path)
                post = path_item.post
                if post:
                    req_model = getattr(models, post.requestBodyModel, EmptyBody)
                    route_info = RouteInfo(
                        request_model=req_model, description=post.description
                    )
                    if post.responseModels and post.responseModels.get(200):
                        resp_model = getattr(models, post.responseModels[200])
                        route_info.response_model = resp_model
                    self._routes.post_map[path] = route_info

    def get_response_model(
        self, path: str, method: str
    ) -> Optional[Type[pydantic.BaseModel]]:
        """
        Get response model for a specific path
        :param path: Path of the route, e.g "/pets"
        :param method: HTTP Method, e.g "post" or "GET"
        :return: Pydantic model with the fields defined in spec
        """
        store = getattr(self._routes, f"{method.lower()}_map", None)
        if store is None:
            raise ValueError("Unsupported HTTP method")
        route_info: RouteInfo = store.get(path)
        if not route_info:
            return None
        return route_info.response_model

    def post(
        self,
        path: Optional[str] = None,
        name: str = None,
        tags: List[str] = None,
        description: str = None,
        response_description: str = None,
        name_factory: Optional[Callable] = None,
    ):
        """
        Define implementation for a specific POST route
        If path argument is not provided, then this handler will be used in all POST
        routes found in the OpenAPI spec
        :param path: Path of the route, e.g "/pets"
        :param name: Name of the route, mainly used in docs
        :param name_factory: Function to generate route name from path.
            It has precedence over `name` param
        :param tags: Specific tags for docs
        :param description: Route description. Got from OpenAPI spec by default
        :param response_description: Description of the response
        """

        def _wrapper(fn):
            if path is None:
                route_info = RouteInfo()
                self._routes.default_post = route_info
            else:
                route_info = self._routes.post_map[path]

            route_info.handler = fn
            route_info.name = name
            route_info.tags = tags
            route_info.name_factory = name_factory
            if response_description:
                route_info.response_description = response_description
            if description:
                route_info.description = description

            return fn

        return _wrapper

    def to_fastapi_router(self):
        """
        Creates an instance of FastAPI router. Must be called after route definitions
        :return: APIRouter instance
        """
        router = APIRouter()

        # POST methods
        for path, route_info in self._routes.post_map.items():
            resp_model = self.get_response_model(path, "post")
            req_model = route_info.request_model

            # if route is not customized, fall back to default POST
            if route_info.handler == dummy_route:
                route_info = self._routes.default_post

            route_name = route_info.name
            if route_info.name_factory:
                route_name = route_info.name_factory(path)

            handler = copy_function(route_info.handler)
            add_annotation_to_first_argument(handler, req_model)  # noqa type: ignore

            router.post(
                path,
                name=route_name,
                description=route_info.description,
                response_description=route_info.response_description,
                response_model=resp_model,
                tags=route_info.tags,
            )(handler)
        return router



from pathlib import Path

import pytest
from fastapi import FastAPI
from starlette.testclient import TestClient

from openapi_to_fastapi.routes import SpecRouter


@pytest.fixture
def specs_root():
    return (Path(__file__).absolute().parent / "data").absolute()


@pytest.fixture
def app():
    return FastAPI()


@pytest.fixture
def client(app):
    return TestClient(app)


@pytest.fixture
def ihan_client(specs_root):
    app = FastAPI()
    spec_router = SpecRouter(specs_root / "ihan")
    app.include_router(spec_router.to_fastapi_router())
    return TestClient(app)



"""
This module extends some of FastAPI models, to include extra data into them
"""

from typing import Dict, Optional

from fastapi.openapi import models as oas


class Operation(oas.Operation):
    requestBodyModel: Optional[str] = ""
    responseModels: Optional[Dict[int, str]] = None


class PathItem(oas.PathItem):
    post: Optional[Operation] = None
    get: Optional[Operation] = None



from typing import Dict, List, Optional

from fastapi.openapi import models as oas

from .models import Operation, PathItem
from .validator import MissingParameter


def parse_parameters(spec: dict) -> List[oas.Parameter]:
    """
    Create Parameter structure from the OpenAPI spec. Loads what kind of parameters
    are used in the request and which of them are required
    :param spec: Part of an OpenAPI spec for a certain route
    :return: List of Parameter objects
    """
    parameters = []
    for param in spec.get("parameters", []):
        try:
            param_name = param["name"]
        except KeyError:
            raise MissingParameter(f"Name is missing: {param}")
        param_in_ = oas.ParameterInType(param["in"])
        parameter_item = oas.Parameter(**{"name": param_name, "in": param_in_})
        parameter_item.required = param.get("required", False)
        parameter_item.description = param.get("description")
        parameters.append(parameter_item)

    return parameters


def get_model_name_from_ref(spec: dict) -> Optional[str]:
    """
    Get the name of a model schema, which defines a request body or a response.
    Currently only application/json schemas are supported

    :param spec: Part of OpenAPI spec, which defines a model. For example,
    `/paths/<route>/post/requestBody` or `paths/<route>/get/responses/200`
    :return: Model name, for example `BasicCompanyInfoResponse`
    """
    app_json = spec.get("content", {}).get("application/json", {})
    if not app_json:
        return None

    ref_name = app_json.get("schema", {}).get("$ref")
    if ref_name:
        return str(ref_name).split("/")[-1]
    return None


def parse_operation(spec: dict, name: str) -> Optional[Operation]:
    """
    Create Operation structure from the OpenAPI spec. Loads meta information
    like what model defines request/response body of an endpoint
    :param spec: Part of an OpenAPI spec for a certain route
    :param name: Operation name, like "post" or "get"
    :return: Operation item, if it exists
    """

    if not spec.get(name):
        return None

    data = spec[name]
    operation = Operation(responses={}, responseModels={})
    operation.parameters = parse_parameters(spec[name])  # type: ignore

    if data.get("description"):
        operation.description = data.get("description")

    request_body_model = get_model_name_from_ref(data.get("requestBody", {}))
    if request_body_model:
        operation.requestBodyModel = request_body_model

    operation.responseModels = {}
    for resp_code, resp_data in data.get("responses", {}).items():
        code = int(resp_code)
        model_name = get_model_name_from_ref(resp_data)
        if model_name:
            operation.responseModels[code] = model_name

    return operation


def parse_openapi_spec(spec: dict) -> Dict[str, PathItem]:
    """
    Create PathItem structures from the OpenAPI spec. It contains required
    information for setting up routing

    :param spec: OpenAPI spec as dictionary
    :return: Mapping from path to PathItem
    """
    path_items: Dict[str, PathItem] = {}
    for path, data in spec.get("paths", {}).items():
        path_item = PathItem(description=data.get("description"))

        post_operation = parse_operation(data, "post")
        if post_operation:
            path_item.post = post_operation

        get_operation = parse_operation(data, "get")
        if get_operation:
            path_item.get = get_operation

        path_items[path] = path_item

    return path_items



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['openapi_to_fastapi',
 'openapi_to_fastapi.tests',
 'openapi_to_fastapi.tests.snapshots',
 'openapi_to_fastapi.validator']

package_data = \
{'': ['*'], 'openapi_to_fastapi.tests': ['data/ihan/*', 'data/pet_store/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'coloredlogs>=14.0,<15.0',
 'datamodel-code-generator>=0.5.24,<0.6.0',
 'fastapi>=0.47.1,<0.48.0',
 'genson==1.2.1']

entry_points = \
{'console_scripts': ['openapi-validator = '
                     'openapi_to_fastapi.cli:cli_validate_specs']}

setup_kwargs = {
    'name': 'openapi-to-fastapi',
    'version': '0.1.2',
    'description': 'Create FastAPI routes from OpenAPI spec',
    'long_description': '## Reasoning\n\n[FastAPI](https://github.com/tiangolo/fastapi) is an awesome framework that\nsimplifies the process of creating APIs. One of the most exciting features is\nthat it can generate OpenAPI specs out of the box. But what if.. you have an\nOpenAPI spec and you need to create an API from it?\n\nOne day we faced that problem — we had to create an API from multiple OpenAPI\nspecs, and make sure that the incoming requests and the outgoing responses were\naligned with the models defined the specs.\n\n> ⚠️ This library was created to cover only our own needs first. So for now it\'s\n> not suitable for everyone and has a lot of technical restrictions. Please\n> consider it as experimental stuff\n\n## Installation\n\nThe package is available on PyPi:\n\n```bash\npip install openapi-to-fastapi\n```\n\n## Generating FastAPI routes\n\nThe main purpose of this library is to generate FastAPI routes from OpenAPI\nspecs. This is done by:\n\n```python\nfrom pathlib import Path\nfrom openapi_to_fastapi.routes import SpecRouter\n\nspecs = Path("./specs")\n\nrouter = SpecRouter(specs).to_fastapi_router()\n```\n\nThe code above will create a FastAPI router that can be either included into the\nmain router, or used as the default one.\n\nImagine you have a following spec (some parts are cut off for brevity):\n\n```json\n{\n  "openapi": "3.0.2",\n  "paths": {\n    "/Company/BasicInfo": {\n      "post": {\n        "requestBody": {\n          "required": true,\n          "content": {\n            "application/json": {\n              "schema": {\n                "$ref": "#/components/schemas/BasicCompanyInfoRequest",\n                "responses": {\n                  "200": {\n                    "content": {\n                      "application/json": {\n                        "schema": {\n                          "$ref": "#/components/schemas/BasicCompanyInfoResponse"\n                        }\n                      }\n                    }\n                  }\n                }\n              }\n            }\n          }\n        }\n      },\n      "components": {\n        "schemas": {\n          "BasicCompanyInfoRequest": {\n            "title": "BasicCompanyInfoRequest",\n            "required": ["companyId"],\n            "type": "object",\n            "properties": {\n              "companyId": {\n                "title": "Company Id",\n                "type": "string",\n                "example": "2464491-9"\n              }\n            }\n          },\n          "BasicCompanyInfoResponse": {\n            "title": "BasicCompanyInfoResponse",\n            "required": ["name", "companyId", "companyForm"],\n            "type": "object",\n            "properties": {\n              "name": {\n                "title": "Name of the company",\n                "type": "string"\n              },\n              "companyId": {\n                "title": "ID of the company",\n                "type": "string"\n              },\n              "companyForm": {\n                "title": "The company form of the company",\n                "type": "string"\n              }\n            }\n          }\n        }\n      }\n    }\n  }\n}\n```\n\nThe FastAPI route equivalent could look like this:\n\n```python\nclass BasicCompanyInfoRequest(pydantic.BaseModel):\n    companyId: str\n\nclass BasicCompanyInfoResponse(pydantic.BaseModel):\n    name: str\n    companyId: str\n    companyForm: str\n\n\n@router.post("/Company/BasicInfo", response_model=BasicCompanyInfoResponse)\ndef _route(request: BasicCompanyInfoRequest):\n    return {}\n\n```\n\nAnd `openapi-to-fastapi` can create it automagically.\n\n### Custom routes\n\nIn most cases it makes no sense to create an API without any business logic.\n\nHere\'s how to define it:\n\n```python\nfrom fastapi import Header, HTTPException\nfrom openapi_to_fastapi.routes import SpecRouter\n\nspec_router = SpecRouter("./specs")\n\n# Default handler for all POST endpoints found in the spec\n@spec_router.post()\ndef hello_world(params, x_my_token: str = Header(...)):\n    if x_my_token != "my_token":\n        raise HTTPException(status_code=403, detail="Sorry")\n    return {"Hello": "World"}\n\n# Specific endpoint for a "/pet" route\n@spec_router.post("/pet")\ndef create_pet(params):\n        pet = db.make_pet(name=params.name)\n        return pet.to_dict()\n\nrouter = spec_router.to_fastapi_router()\n```\n\n### API Documentation\n\nNow after you have a lot of routes, you might want to leverage another great\nfeature of FastAPI — auto documentation.\n\nRequest and response models are already handled. But to display documentation\nnicely, FastAPI needs to assign a name for each endpoint. Here is how you can\nprovide such name:\n\n```python\nfrom openapi_to_fastapi.routes import SpecRouter\n\nspec_router = SpecRouter("./specs")\n\n@spec_router.post(\n    "/pet",\n    name="Create a pet",\n    description="Create a pet",\n    response_description="A Pet",\n    tags=["pets"],\n)\ndef create_pet(params):\n    return {}\n\n# Or you can set the dynamic name based on API path\ndef name_factory(path: str, **kwargs):\n    return path.replace("/", " ")\n\n@spec_router.post(name_factory=name_factory)\ndef create_pet(params):\n    return {}\n\n```\n\n## OpenAPI validation\n\nThis package also provides a CLI entrypoint to validate OpenAPI specs. It\'s\nespecially useful when you need to define you own set of rules for validation.\n\nImagine your API specs are stored in a separate repository and maintained by\nanother team. You also expect that all OpenAPI specs have only one endpoint\ndefined (some internal agreement).\n\nNow you can set up a CI check and validate them on every push.\n\nFirstly create a file with a custom validator:\n\n```python\n# my_validator.py\n\nfrom openapi_to_fastapi.validator import BaseValidator, OpenApiValidationError\n\nclass CustomError(OpenApiValidationError):\n    pass\n\n# important: must be inherited from BaseValidator\nclass MyValidator(BaseValidator):\n\n    # implement this single method\n    def validate_spec(self, spec: dict):\n        if len(spec["paths"]) != 1:\n            raise CustomError("Only one endpoint allowed")\n```\n\nThen run the tool:\n\n```\nopenapi-validator --path ./standards -m my_validator.py -v MyValidator\n\n===============================================================================\nOpenAPI specs root path: ./standards\nValidators: DefaultValidator, MyValidator\n===============================================================================\nFile: ./standards/Current.json\n[PASSED]\n-------------------------------------------------------------------------------\nFile: ./standards/Metric.json\n[PASSED]\n-------------------------------------------------------------------------------\nFile: ./standards/BasicInfo.json\n[PASSED]\n-------------------------------------------------------------------------------\n===============================================================================\nSummary:\nTotal : 3\nPassed: 3\nFailed: 0\n===============================================================================\n```\n\nThis validator can also be reused when generating routes:\n\n```python\nrouter = SpecRouter(specs, validators=[MyValidator])\n```\n\n## License\n\nThis code is released under the BSD 3-Clause license. Details in the\n[LICENSE](./LICENSE) file.\n',
    'author': 'Digital Living International Ltd',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/digitalliving/openapi-to-fastapi',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)
